@@ -9,7 +9,6 @@ but they're not necessarily the most performant.
 from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
-from __future__ import unicode_literals
 
 import codecs
 import datetime
@@ -22,19 +21,27 @@ import pandas as pd
 import struct
 import sys
 from decimal import Decimal
+import binascii
 
 from .thrift_structures import parquet_thrift
+from .util import PY2
+from .speedups import array_decode_utf8
+
 logger = logging.getLogger('parquet')  # pylint: disable=invalid-name
 
 try:
     from bson import BSON
-    unbson = lambda s: BSON(s).decode()
+    unbson = BSON.decode
+    tobson = BSON.encode
 except ImportError:
     try:
         import bson
         unbson = bson.loads
+        tobson = bson.dumps
     except:
-        def BSON(x):
+        def unbson(x):
+            raise ImportError("BSON not found")
+        def tobson(x):
             raise ImportError("BSON not found")
 
 DAYS_TO_MILLIS = 86400000000000
@@ -79,28 +86,41 @@ def typemap(se):
     return np.dtype("O")
 
 
-def convert(data, se):
+def convert(data, se, timestamp96=False):
     """Convert known types from primitive to rich.
 
     Parameters
     ----------
     data: pandas series of primitive type
     se: a schema element.
+    timestamp96: convert int96 as if it were written by mr-parquet
     """
-    #data = np.asarray(data, dtype=simple[se.type])
     ctype = se.converted_type
+    if se.type == parquet_thrift.Type.INT96 and timestamp96:
+        data2 = data.view([('ns', 'i8'), ('day', 'i4')])
+        return ((data2['day'] - 2440588) * 86400000000000 +
+                data2['ns']).view('M8[ns]')
     if ctype is None:
         return data
     if ctype == parquet_thrift.ConvertedType.UTF8:
         if isinstance(data, list) or data.dtype != "O":
-            out = np.empty(len(data), dtype="O")
-        else:
-            out = data
-        out[:] = [s.decode('utf8') for s in data]
-        return out
+            data = np.asarray(data, dtype="O")
+        return array_decode_utf8(data)
     if ctype == parquet_thrift.ConvertedType.DECIMAL:
         scale_factor = 10**-se.scale
-        return data * scale_factor
+        if data.dtype.kind in ['i', 'f']:
+            return data * scale_factor
+        else:  # byte-string
+            # NB: general but slow method
+            # could optimize when data.dtype.itemsize <= 8
+            if PY2:
+                def from_bytes(d):
+                    return int(binascii.b2a_hex(d), 16) if len(d) else 0
+                return np.array([from_bytes(d) * scale_factor for d in data])
+            else:
+                # NB: `from_bytes` may be py>=3.4 only
+                return np.array([int.from_bytes(d, byteorder='big', signed=True) *
+                                 scale_factor for d in data])
     elif ctype == parquet_thrift.ConvertedType.DATE:
         return (data * DAYS_TO_MILLIS).view('datetime64[ns]')
     elif ctype == parquet_thrift.ConvertedType.TIME_MILLIS:
@@ -127,6 +147,14 @@ def convert(data, se):
         return data.astype(np.uint32)
     elif ctype == parquet_thrift.ConvertedType.UINT_64:
         return data.astype(np.uint64)
+    elif ctype == parquet_thrift.ConvertedType.INT_8:
+        return data.astype(np.int8)
+    elif ctype == parquet_thrift.ConvertedType.INT_16:
+        return data.astype(np.int16)
+    elif ctype == parquet_thrift.ConvertedType.INT_32:
+        return data.astype(np.int32)
+    elif ctype == parquet_thrift.ConvertedType.INT_64:
+        return data.astype(np.int64)
     elif ctype == parquet_thrift.ConvertedType.JSON:
         if isinstance(data, list) or data.dtype != "O":
             out = np.empty(len(data), dtype="O")
@@ -151,8 +179,8 @@ def convert(data, se):
     return data
 
 
-@numba.njit()
-def time_shift(indata, outdata, factor=1000):
+@numba.njit(nogil=True)
+def time_shift(indata, outdata, factor=1000):  # pragma: no cover
     for i in range(len(indata)):
         if indata[i] == nat:
             outdata[i] = nat
