@@ -5,27 +5,28 @@ import struct
 
 import numpy as np
 import pandas as pd
-from thriftpy.protocol.compact import TCompactProtocolFactory
+from thriftpy.protocol.compact import TCompactProtocol
 
 from . import encoding
 from .compression import decompress_data
 from .converted_types import convert, typemap
-from .thrift_filetransport import TFileTransport
+from .schema import _is_list_like, _is_map_like
+from .speedups import unpack_byte_array
 from .thrift_structures import parquet_thrift
-from .util import val_to_num
+from .util import val_to_num, byte_buffer, ex_from_sep
 
 
 def read_thrift(file_obj, ttype):
     """Read a thrift structure from the given fo."""
-    tin = TFileTransport(file_obj)
-    pin = TCompactProtocolFactory().get_protocol(tin)
+    pin = TCompactProtocol(file_obj, True)
     page_header = ttype()
     page_header.read(pin)
     return page_header
 
 
 def _read_page(file_obj, page_header, column_metadata):
-    """Read the data page from the given file-object and convert it to raw, uncompressed bytes (if necessary)."""
+    """Read the data page from the given file-object and convert it to raw, 
+    uncompressed bytes (if necessary)."""
     raw_bytes = file_obj.read(page_header.compressed_page_size)
     raw_bytes = decompress_data(raw_bytes, column_metadata.codec)
 
@@ -57,7 +58,7 @@ def read_def(io_obj, daph, helper, metadata):
     """
     definition_levels = None
     num_nulls = 0
-    if not helper.is_required(metadata.path_in_schema[-1]):
+    if not helper.is_required(metadata.path_in_schema):
         max_definition_level = helper.max_definition_level(
             metadata.path_in_schema)
         bit_width = encoding.width_from_max_int(max_definition_level)
@@ -65,8 +66,8 @@ def read_def(io_obj, daph, helper, metadata):
             definition_levels = read_data(
                     io_obj, daph.definition_level_encoding,
                     daph.num_values, bit_width)[:daph.num_values]
-        num_nulls = daph.num_values - (definition_levels ==
-                                       max_definition_level).sum()
+            num_nulls = daph.num_values - (definition_levels ==
+                                           max_definition_level).sum()
         if num_nulls == 0:
             definition_levels = None
     return definition_levels, num_nulls
@@ -76,7 +77,7 @@ def read_rep(io_obj, daph, helper, metadata):
     """
     Read the repetition levels from this page, if any.
     """
-    repetition_levels = None  # pylint: disable=unused-variable
+    repetition_levels = None
     if len(metadata.path_in_schema) > 1:
         max_repetition_level = helper.max_repetition_level(
             metadata.path_in_schema)
@@ -84,10 +85,13 @@ def read_rep(io_obj, daph, helper, metadata):
         repetition_levels = read_data(io_obj, daph.repetition_level_encoding,
                                       daph.num_values,
                                       bit_width)[:daph.num_values]
+        if repetition_levels.max() == 0:
+            repetition_levels = None
     return repetition_levels
 
 
-def read_data_page(f, helper, header, metadata, skip_nulls=False):
+def read_data_page(f, helper, header, metadata, skip_nulls=False,
+                   selfmade=False):
     """Read a data page: definitions, repetitions, values (in order)
 
     Only values are guaranteed to exist, e.g., for a top-level, required
@@ -95,38 +99,48 @@ def read_data_page(f, helper, header, metadata, skip_nulls=False):
     """
     daph = header.data_page_header
     raw_bytes = _read_page(f, header, metadata)
-    io_obj = encoding.Numpy8(np.frombuffer(memoryview(raw_bytes),
+    io_obj = encoding.Numpy8(np.frombuffer(byte_buffer(raw_bytes),
                                            dtype=np.uint8))
 
-    if skip_nulls and not helper.is_required(metadata.path_in_schema[-1]):
+    repetition_levels = read_rep(io_obj, daph, helper, metadata)
+
+    if skip_nulls and not helper.is_required(metadata.path_in_schema):
         num_nulls = 0
         definition_levels = None
         skip_definition_bytes(io_obj, daph.num_values)
     else:
         definition_levels, num_nulls = read_def(io_obj, daph, helper, metadata)
 
-    repetition_levels = read_rep(io_obj, daph, helper, metadata)
+    nval = daph.num_values - num_nulls
     if daph.encoding == parquet_thrift.Encoding.PLAIN:
-        width = helper.schema_element(metadata.path_in_schema[-1]).type_length
+        width = helper.schema_element(metadata.path_in_schema).type_length
         values = encoding.read_plain(raw_bytes[io_obj.loc:],
                                      metadata.type,
                                      int(daph.num_values - num_nulls),
                                      width=width)
-    elif daph.encoding == parquet_thrift.Encoding.PLAIN_DICTIONARY:
+    elif daph.encoding in [parquet_thrift.Encoding.PLAIN_DICTIONARY,
+                           parquet_thrift.Encoding.RLE]:
         # bit_width is stored as single byte.
-        bit_width = io_obj.read_byte()
-        if bit_width:
-            values = encoding.Numpy32(np.zeros(daph.num_values,
+        if daph.encoding == parquet_thrift.Encoding.RLE:
+            bit_width = helper.schema_element(
+                    metadata.path_in_schema).type_length
+        else:
+            bit_width = io_obj.read_byte()
+        if bit_width in [8, 16, 32] and selfmade:
+            num = (encoding.read_unsigned_var_int(io_obj) >> 1) * 8
+            values = io_obj.read(num * bit_width // 8).view('int%i' % bit_width)
+        elif bit_width:
+            values = encoding.Numpy32(np.empty(daph.num_values-num_nulls+7,
                                                dtype=np.int32))
             # length is simply "all data left in this page"
             encoding.read_rle_bit_packed_hybrid(
                         io_obj, bit_width, io_obj.len-io_obj.loc, o=values)
-            values = values.data[:daph.num_values-num_nulls]
+            values = values.data[:nval]
         else:
-            values = np.zeros(daph.num_values-num_nulls, dtype=np.int64)
+            values = np.zeros(nval, dtype=np.int8)
     else:
         raise NotImplementedError('Encoding %s' % daph.encoding)
-    return definition_levels, repetition_levels, values
+    return definition_levels, repetition_levels, values[:nval]
 
 
 def skip_definition_bytes(io_obj, num):
@@ -144,13 +158,11 @@ def read_dictionary_page(file_obj, schema_helper, page_header, column_metadata):
     """
     raw_bytes = _read_page(file_obj, page_header, column_metadata)
     if column_metadata.type == parquet_thrift.Type.BYTE_ARRAY:
-        # no faster way to read variable-length-strings?
-        fobj = io.BytesIO(raw_bytes)
-        values = [fobj.read(struct.unpack(b"<i", fobj.read(4))[0])
-                  for _ in range(page_header.dictionary_page_header.num_values)]
+        values = unpack_byte_array(raw_bytes,
+                                   page_header.dictionary_page_header.num_values)
     else:
         width = schema_helper.schema_element(
-            column_metadata.path_in_schema[-1]).type_length
+            column_metadata.path_in_schema).type_length
         values = encoding.read_plain(
                 raw_bytes, column_metadata.type,
                 page_header.dictionary_page_header.num_values, width)
@@ -158,7 +170,8 @@ def read_dictionary_page(file_obj, schema_helper, page_header, column_metadata):
 
 
 def read_col(column, schema_helper, infile, use_cat=False,
-             grab_dict=False, selfmade=False):
+             grab_dict=False, selfmade=False, assign=None, catdef=None,
+             timestamp96=False):
     """Using the given metadata, read one column in one row-group.
 
     Parameters
@@ -175,9 +188,12 @@ def read_col(column, schema_helper, infile, use_cat=False,
     grab_dict: bool (False)
         Short-cut mode to return the dictionary values only - skips the actual
         data.
+    timestamp96: bool
+        If True, and if this is an int96 field, interpret as a timestamp as
+        used in MapReduce-based tools.
     """
     cmd = column.meta_data
-    se = schema_helper.schema_element(cmd.path_in_schema[-1])
+    se = schema_helper.schema_element(cmd.path_in_schema)
     off = min((cmd.dictionary_page_offset or cmd.data_page_offset,
                cmd.data_page_offset))
 
@@ -188,90 +204,90 @@ def read_col(column, schema_helper, infile, use_cat=False,
     if ph.type == parquet_thrift.PageType.DICTIONARY_PAGE:
         dic = np.array(read_dictionary_page(infile, schema_helper, ph, cmd))
         ph = read_thrift(infile, parquet_thrift.PageHeader)
-        dic = convert(dic, se)
+        dic = convert(dic, se, timestamp96=timestamp96)
     if grab_dict:
         return dic
+    if use_cat:
+        catdef._categories = pd.Index(dic)
+        if np.iinfo(assign.dtype).max < len(dic):
+            raise RuntimeError('Assigned array dtype (%s) cannot accommodate '
+                               'number of category labels (%i)' %
+                               (assign.dtype, len(dic)))
 
     rows = cmd.num_values
 
-    out = []
+    do_convert = True
+    if use_cat:
+        my_nan = -1
+        do_convert = False
+    else:
+        if assign.dtype.kind in ['f', 'i']:
+            my_nan = np.nan
+        elif assign.dtype.kind in ["M", 'm']:
+            my_nan = -9223372036854775808  # int64 version of NaT
+        else:
+            my_nan = None
+
     num = 0
     while True:
-        # TODO: under assumption such as all_dict, could allocate once
-        # and fill arrays, i.e., merge this loop and the next
         if (selfmade and hasattr(cmd, 'statistics') and
                 getattr(cmd.statistics, 'null_count', 1) == 0):
             skip_nulls = True
         else:
             skip_nulls = False
         defi, rep, val = read_data_page(infile, schema_helper, ph, cmd,
-                                        skip_nulls)
+                                        skip_nulls, selfmade=selfmade)
+        if rep is not None and assign.dtype.kind != 'O':  # pragma: no cover
+            # this should never get called
+            raise ValueError('Column contains repeated value, must use object'
+                             'type, but has assumed type: %s' % assign.dtype)
         d = ph.data_page_header.encoding == parquet_thrift.Encoding.PLAIN_DICTIONARY
-        out.append((defi, rep, val, d))
+        if use_cat and not d:
+            raise ValueError('Returning category type requires all chunks to'
+                             'use dictionary encoding; column: %s',
+                             cmd.path_in_schema)
+
+        if rep is not None:
+            null = not schema_helper.is_required(cmd.path_in_schema[0])
+            null_val = (se.repetition_type !=
+                                 parquet_thrift.FieldRepetitionType.REQUIRED)
+            num = encoding._assemble_objects(assign, defi, rep, val, dic, d,
+                                             null, null_val)
+        elif defi is not None:
+            part = assign[num:num+len(defi)]
+            part[defi != 1] = my_nan
+            if d and not use_cat:
+                part[defi == 1] = dic[val]
+            elif do_convert:
+                part[defi == 1] = convert(val, se, timestamp96=timestamp96)
+            else:
+                part[defi == 1] = val
+        else:
+            piece = assign[num:num+len(val)]
+            if d and not use_cat:
+                piece[:] = dic[val]
+            elif do_convert:
+                piece[:] = convert(val, se, timestamp96=timestamp96)
+            else:
+                piece[:] = val
+
         num += len(defi) if defi is not None else len(val)
         if num >= rows:
             break
         ph = read_thrift(infile, parquet_thrift.PageHeader)
 
-    all_dict = use_cat and all(_[3] for _ in out)
-    any_def = any(_[0] is not None for _ in out)
-    do_convert = True
-    if all_dict:
-        dtype = np.int64
-        my_nan = -1
-        do_convert = False
-    else:
-        dtype = typemap(se)  # output dtype
-        if any_def and dtype.kind == 'i':
-            # integers cannot hold NULLs/NaNs
-            dtype = np.dtype('float64')
-            do_convert = False
-        if dtype.kind == 'f':
-            my_nan = np.nan
-        elif dtype.kind in ["M", 'm']:
-            my_nan = -9223372036854775808  # int64 version of NaT
-        else:
-            my_nan = None
-    if len(out) == 1 and not any_def:
-        defi, rep, val, d = out[0]
-        if d and not all_dict:
-            final = dic[val]
-        elif do_convert:
-            final = convert(val, se)
-        else:
-            final = val
-    else:
-        final = np.empty(cmd.num_values, dtype)
-        start = 0
-        for defi, rep, val, d in out:
-            if d and not all_dict:
-                cval = dic[val]
-            elif do_convert:
-                cval = convert(val, se)
-            else:
-                cval = val
-            if defi is not None:
-                part = final[start:start+len(defi)]
-                part[defi != 1] = my_nan
-                part[defi == 1] = cval
-                start += len(defi)
-            else:
-                final[start:start+len(val)] = cval
-                start += len(val)
-    if all_dict:
-        final = pd.Categorical.from_codes(final, categories=dic)
-    return final
 
-
-def read_row_group_file(fn, columns, *args, open=open, selfmade=False,
-                        index=None):
+def read_row_group_file(fn, rg, columns, categories, schema_helper, cats,
+                        open=open, selfmade=False, index=None, assign=None,
+                        timestamp96=[], sep=os.sep):
     with open(fn, mode='rb') as f:
-        return read_row_group(f, columns, *args, selfmade=selfmade,
-                              index=index)
+        return read_row_group(f, rg, columns, categories, schema_helper, cats,
+                              selfmade=selfmade, index=index, assign=assign,
+                              timestamp96=timestamp96, sep=sep)
 
 
 def read_row_group_arrays(file, rg, columns, categories, schema_helper, cats,
-                          selfmade=False):
+                          selfmade=False, assign=None, timestamp96=[]):
     """
     Read a row group and return as a dict of arrays
 
@@ -279,44 +295,53 @@ def read_row_group_arrays(file, rg, columns, categories, schema_helper, cats,
     will be pandas Categorical objects: the codes and the category labels
     are arrays.
     """
-    out = {}
+    out = assign
+    maps = {}
 
     for column in rg.columns:
         name = ".".join(column.meta_data.path_in_schema)
-        if name not in columns:
+        if (column.meta_data.path_in_schema[0] in columns and
+                (_is_list_like(schema_helper, column) or
+                 _is_map_like(schema_helper, column))):
+            name = column.meta_data.path_in_schema[0]
+        elif name not in columns:
             continue
 
         use = name in categories if categories is not None else False
-        s = read_col(column, schema_helper, file, use_cat=use,
-                     selfmade=selfmade)
-        out[name] = s
-    return out
+        mr = name in timestamp96
+        read_col(column, schema_helper, file, use_cat=use,
+                 selfmade=selfmade, assign=out[name],
+                 catdef=out[name+'-catdef'] if use else None,
+                 timestamp96=mr)
+
+        if _is_map_like(schema_helper, column):
+            if name not in maps:
+                maps[name] = out[name].copy()
+            else:
+                if column.meta_data.path_in_schema[0] == 'key':
+                    key, value = out[name], maps[name]
+                else:
+                    value, key = out[name], maps[name]
+                for i, (k, v) in enumerate(zip(key, value)):
+                    out[name][i] = dict(zip(k, v)) if k is not None else None
 
 
 def read_row_group(file, rg, columns, categories, schema_helper, cats,
-                   selfmade=False, index=None):
+                   selfmade=False, index=None, assign=None, timestamp96=[],
+                   sep=os.sep):
     """
     Access row-group in a file and read some columns into a data-frame.
     """
-    out = read_row_group_arrays(file, rg, columns, categories, schema_helper,
-                                cats, selfmade)
+    if assign is None:
+        raise RuntimeError('Going with pre-allocation!')
+    read_row_group_arrays(file, rg, columns, categories, schema_helper,
+                          cats, selfmade, assign=assign, timestamp96=timestamp96)
 
-    if index is not None and index in columns:
-        i = out.pop(index)
-        out = pd.DataFrame(out, index=i)
-        out.index.name = index
-    else:
-        out = pd.DataFrame(out, columns=columns)
-
-    # apply categories
     for cat in cats:
-        # *Hard assumption*: all chunks in a row group have the
-        # same partition (correct for spark/hive)
-        partitions = re.findall("([a-zA-Z_]+)=([^/]+)/",
-                                rg.columns[0].file_path)
-        val = [p[1] for p in partitions if p[0] == cat][0]
-        codes = np.empty(rg.num_rows, dtype=np.int16)
-        codes[:] = cats[cat].index(val)
-        out[cat] = pd.Categorical.from_codes(
-                codes, [val_to_num(c) for c in cats[cat]])
-    return out
+        s = ex_from_sep(sep)
+        partitions = s.findall(rg.columns[0].file_path)
+        if not partitions and sep in (rg.columns[0].file_path or ""):
+            partitions = [('dir%i' % i, v) for (i, v) in enumerate(
+                rg.columns[0].file_path.split(sep)[:-1])]
+        val = val_to_num([p[1] for p in partitions if p[0] == cat][0])
+        assign[cat][:] = cats[cat].index(val)
