@@ -13,7 +13,9 @@ import os
 import struct
 import sys
 
+from .speedups import unpack_byte_array
 from .thrift_structures import parquet_thrift
+from .util import byte_buffer
 
 
 def read_plain_boolean(raw_bytes, count):
@@ -33,22 +35,27 @@ DECODE_TYPEMAP = {
 def read_plain(raw_bytes, type_, count, width=0):
     if type_ in DECODE_TYPEMAP:
         dtype = DECODE_TYPEMAP[type_]
-        return np.frombuffer(memoryview(raw_bytes), dtype=dtype, count=count)
+        return np.frombuffer(byte_buffer(raw_bytes), dtype=dtype, count=count)
     if type_ == parquet_thrift.Type.FIXED_LEN_BYTE_ARRAY:
         if count == 1:
             width = len(raw_bytes)
         dtype = np.dtype('S%i' % width)
-        return np.frombuffer(memoryview(raw_bytes), dtype=dtype, count=count)
+        return np.frombuffer(byte_buffer(raw_bytes), dtype=dtype, count=count)
     if type_ == parquet_thrift.Type.BOOLEAN:
         return read_plain_boolean(raw_bytes, count)
     # variable byte arrays (rare)
-    file_obj = io.BytesIO(raw_bytes)
-    return np.array([file_obj.read(struct.unpack('<i', file_obj.read(4))[0])
-                     for _ in range(count)], dtype="O")
+    try:
+        return np.array(unpack_byte_array(raw_bytes, count), dtype='O')
+    except RuntimeError:
+        if count == 1:
+            # e.g., for statistics
+            return np.array([raw_bytes], dtype='O')
+        else:
+            raise
 
 
 @numba.jit(nogil=True)
-def read_unsigned_var_int(file_obj):
+def read_unsigned_var_int(file_obj):  # pragma: no cover
     """Read a value using the unsigned, variable int encoding.
     file-obj is a NumpyIO of bytes; avoids struct to allow numba-jit
     """
@@ -64,7 +71,7 @@ def read_unsigned_var_int(file_obj):
 
 
 @numba.njit(nogil=True)
-def read_rle(file_obj, header, bit_width, o):
+def read_rle(file_obj, header, bit_width, o):  # pragma: no cover
     """Read a run-length encoded run from the given fo with the given header and bit_width.
 
     The count is determined from the header and the width is used to grab the
@@ -80,7 +87,7 @@ def read_rle(file_obj, header, bit_width, o):
 
 
 @numba.njit(nogil=True)
-def width_from_max_int(value):
+def width_from_max_int(value):  # pragma: no cover
     """Convert the value specified to a bit_width."""
     for i in range(0, 64):
         if value == 0:
@@ -89,13 +96,13 @@ def width_from_max_int(value):
 
 
 @numba.njit(nogil=True)
-def _mask_for_bits(i):
+def _mask_for_bits(i):  # pragma: no cover
     """Generate a mask to grab `i` bits from an int value."""
     return (1 << i) - 1
 
 
 @numba.njit(nogil=True)
-def read_bitpacked(file_obj, header, width, o):
+def read_bitpacked(file_obj, header, width, o):  # pragma: no cover
     """
     Read values packed into width-bits each (which can be >8)
 
@@ -129,7 +136,7 @@ def read_bitpacked(file_obj, header, width, o):
 
 
 @numba.njit(nogil=True)
-def read_rle_bit_packed_hybrid(io_obj, width, length=None, o=None):
+def read_rle_bit_packed_hybrid(io_obj, width, length=None, o=None):  # pragma: no cover
     """Read values from `io_obj` using the rel/bit-packed hybrid encoding.
 
     If length is not specified, then a 32-bit int is read first to grab the
@@ -146,11 +153,12 @@ def read_rle_bit_packed_hybrid(io_obj, width, length=None, o=None):
             read_rle(io_obj, header, width, o)
         else:
             read_bitpacked(io_obj, header, width, o)
+    io_obj.loc = start + length
     return o.so_far()
 
 
 @numba.njit(nogil=True)
-def read_length(file_obj):
+def read_length(file_obj):  # pragma: no cover
     """ Numpy trick to get a 32-bit length from four bytes
 
     Equivalent to struct.unpack('<i'), but suitable for numba-jit
@@ -159,7 +167,7 @@ def read_length(file_obj):
     return sub[0] + sub[1]*256 + sub[2]*256*256 + sub[3]*256*256*256
 
 
-class NumpyIO(object):
+class NumpyIO(object):  # pragma: no cover
     """
     Read or write from a numpy arra like a file object
 
@@ -206,3 +214,55 @@ spec8 = [('data', numba.uint8[:]), ('loc', numba.int64), ('len', numba.int64)]
 Numpy8 = numba.jitclass(spec8)(NumpyIO)
 spec32 = [('data', numba.uint32[:]), ('loc', numba.int64), ('len', numba.int64)]
 Numpy32 = numba.jitclass(spec32)(NumpyIO)
+
+
+def _assemble_objects(assign, defi, rep, val, dic, d, null, null_val):
+    """Dremel-assembly of arrays of values into lists
+
+    Parameters
+    ----------
+    assign: array dtype O
+        To insert lists into
+    defi: int array
+        Definition levels, max 3
+    rep: int array
+        Repetition levels, max 1
+    dic: array of labels or None
+        Applied if d is True
+    d: bool
+        Whether to dereference dict values
+    null: bool
+        Can an entry be None?
+    null_val: bool
+        can list elements be None
+    """
+    if d:
+        # dereference dict values
+        val = dic[val]
+    i = 0
+    vali = 0
+    part = []
+    started = False
+    append_value = 1 + null + null_val
+    have_null = False
+    for de, re in zip(defi, rep):
+        if not re:
+            # new row - save what we have
+            if started:
+                assign[i] = None if have_null else part
+                part = []
+                i += 1
+            else:
+                # first time: no row to save yet
+                started = True
+        if de == append_value:
+            # append real value to current item
+            part.append(val[vali])
+            vali += 1
+        elif de > null:
+            # append null to current item
+            part.append(None)
+        # next object is None as opposed to an object
+        have_null = de == 0 and null
+    assign[i] = None if have_null else part
+    return i
