@@ -1,4 +1,4 @@
-"""test_read_support.py - unit and integration tests for reading parquet data."""
+"""test_read.py - unit and integration tests for reading parquet data."""
 
 from __future__ import absolute_import
 from __future__ import division
@@ -22,7 +22,7 @@ import pytest
 import fastparquet
 from fastparquet import writer, core
 
-TEST_DATA = "test-data"
+from fastparquet.test.util import sql, s3, TEST_DATA
 
 
 @pytest.yield_fixture()
@@ -42,11 +42,30 @@ def test_header_magic_bytes(tempdir):
         p = fastparquet.ParquetFile(fn, verify=True)
 
 
+@pytest.mark.parametrize("size", [1, 4, 12, 20])
+def test_read_footer_fail(tempdir, size):
+    """Test reading the footer."""
+    import struct
+    fn = os.path.join(TEST_DATA, "nation.impala.parquet")
+    fout = os.path.join(tempdir, "temp.parquet")
+    with open(fn, 'rb') as f1:
+        with open(fout, 'wb') as f2:
+            f1.seek(-8, 2)
+            head_size = struct.unpack('<i', f1.read(4))[0]
+            f1.seek(-(head_size + 8), 2)
+            block = f1.read(head_size)
+            f2.write(b'0' * 25)  # padding
+            f2.write(block[:-size])
+            f2.write(f1.read())
+    with pytest.raises(TypeError):
+        p = fastparquet.ParquetFile(fout)
+
+
 def test_read_footer():
     """Test reading the footer."""
     p = fastparquet.ParquetFile(os.path.join(TEST_DATA, "nation.impala.parquet"))
     snames = {"schema", "n_regionkey", "n_name", "n_nationkey", "n_comment"}
-    assert {s.name for s in p.schema} == snames
+    assert {s.name for s in p._schema} == snames
     assert set(p.columns) == snames - {"schema"}
 
 files = [os.path.join(TEST_DATA, p) for p in
@@ -58,37 +77,6 @@ cols = ["n_nationkey", "n_name", "n_regionkey", "n_comment"]
 expected = pd.read_csv(csvfile, delimiter="|", index_col=0, names=cols)
 
 
-@pytest.yield_fixture()
-def s3():
-    s3fs = pytest.importorskip('s3fs')
-    moto = pytest.importorskip('moto')
-    m = moto.mock_s3()
-    m.start()
-    s3 = s3fs.S3FileSystem()
-    s3.mkdir(TEST_DATA)
-    paths = []
-    for cat, catnum in product(('fred', 'freda'), ('1', '2', '3')):
-        path = os.sep.join([TEST_DATA, 'split', 'cat=' + cat,
-                            'catnum=' + catnum])
-        files = os.listdir(path)
-        for fn in files:
-            full_path = os.path.join(path, fn)
-            s3.put(full_path, full_path)
-            paths.append(full_path)
-    path = os.path.join(TEST_DATA, 'split')
-    files = os.listdir(path)
-    for fn in files:
-        full_path = os.path.join(path, fn)
-        if os.path.isdir(full_path):
-            continue
-        s3.put(full_path, full_path)
-        paths.append(full_path)
-    yield s3
-    for path in paths:
-        s3.rm(path)
-
-
-
 def test_read_s3(s3):
     s3fs = pytest.importorskip('s3fs')
     s3 = s3fs.S3FileSystem()
@@ -168,8 +156,8 @@ def test_dir_partition():
     pf = fastparquet.ParquetFile(os.path.join(TEST_DATA, "split"))
     out = pf.to_pandas()
     for cat, catnum in product(['fred', 'freda'], [1, 2, 3]):
-        assert (df.num[(df.cat==cat) & (df.catnum==catnum)].tolist()) ==\
-               out.num[(out.cat==cat) & (out.catnum==catnum)].tolist()
+        assert (df.num[(df.cat == cat) & (df.catnum == catnum)].tolist()) ==\
+                out.num[(out.cat == cat) & (out.catnum == catnum)].tolist()
     assert out.cat.dtype == 'category'
     assert out.catnum.dtype == 'category'
     assert out.catnum.cat.categories.dtype == 'int64'
@@ -293,7 +281,7 @@ def test_index(tempdir):
         fastparquet.write(tempdir, d2, file_scheme='hive', write_index=True)
         pf = fastparquet.ParquetFile(tempdir)
         out = pf.to_pandas(index=column, categories=['b'])
-        pd.util.testing.assert_frame_equal(out, d2)
+        pd.util.testing.assert_frame_equal(out, d2, check_categorical=False)
 
 
 def test_skip_length():
@@ -304,3 +292,32 @@ def test_skip_length():
         MockIO.loc = 0
         core.skip_definition_bytes(MockIO, num)
         assert len(block) == MockIO.loc
+
+
+def test_timestamp96():
+    pf = fastparquet.ParquetFile(os.path.join(TEST_DATA, 'mr_times.parq'))
+    out = pf.to_pandas(timestamp96=['date_added'])
+    expected = pd.to_datetime(
+            ["2016-08-01 23:08:01", "2016-08-02 23:08:02",
+             "2016-08-03 23:08:03", "2016-08-04 23:08:04",
+             "2016-08-05 23:08:04", "2016-08-06 23:08:05",
+             "2016-08-07 23:08:06", "2016-08-08 23:08:07",
+             "2016-08-09 23:08:08", "2016-08-10 23:08:09"])
+    assert (out['date_added'] == expected).all()
+
+
+def test_bad_catsize(tempdir):
+    df = pd.DataFrame({'a': pd.Categorical([str(i) for i in range(1024)])})
+    fastparquet.write(tempdir, df, file_scheme='hive')
+    pf = fastparquet.ParquetFile(tempdir)
+    assert pf.categories == {'a': 1024}
+    with pytest.raises(RuntimeError):
+        pf.to_pandas(categories={'a': 2})
+
+
+def test_null_sizes(tempdir):
+    df = pd.DataFrame({'a': [True, None], 'b': [3000, np.nan]}, dtype="O")
+    fastparquet.write(tempdir, df, has_nulls=True, file_scheme='hive')
+    pf = fastparquet.ParquetFile(tempdir)
+    assert pf.dtypes['a'] == 'float16'
+    assert pf.dtypes['b'] == 'float64'
